PetInfoFu = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0",
                                           "AceConsole-2.0",
                                           "AceDB-2.0",
                                           "AceEvent-2.0" -- ,
                                           -- "AceHook-2.1"     -- not required  
                                          )

 -- variables
local L          = AceLibrary("AceLocale-2.2"):new("PetInfoFu")
local BS         = AceLibrary("Babble-Spell-2.2")
local BST        = AceLibrary("Babble-SpellTree-2.2")
local crayon     = AceLibrary("Crayon-2.0")
local dewdrop    = AceLibrary("Dewdrop-2.0")
local tablet     = AceLibrary("Tablet-2.0")

-- hunter specific variables
local prevHappiness = 0

-- warlock specific variables
local demonicKnowledgeRank = 0
local demonicKnowledgeTalentTab = 0
local demonicKnowledgeTalentIndex = 0


-- properties
PetInfoFu.clickableTooltip = true
PetInfoFu.canHideText      = true
PetInfoFu.defaultPosition  = "RIGHT"
-- PetInfoFu.hasIcon          = BS:GetSpellIcon("Train Beast")
PetInfoFu.hasIcon          = true
PetInfoFu.hasNoColor       = false

-- register database stuff
PetInfoFu:RegisterDB("PetInfoFuDB")
PetInfoFu:RegisterDefaults("account", {
                                        stables = {},
                                      }
                          )
PetInfoFu:RegisterDefaults("char", {
                                    -- flags
                                    showPetName = true,
                                    showPetLevel = true,
                                    showAttributes = true,
                                    showCombatStats = true,
                                    showSkills = true,
                                    showStableMapNotes = false,
                                    showStabledPets = false,
                                    trackPetHistory = false,
                                    -- lists
                                    hidden = {},
                                    stabledPets = {},
                                    petHistory = {},
                                    beastTraining = {},
                                    netheredDemons = {},
                                   } 
                         )
PetInfoFu:RegisterChatCommand(L["CMD_OPTIONS"], options)

-- menu methods

function PetInfoFu:IsShowingPetName()
	return self.db.char.showPetName
end

function PetInfoFu:ToggleIsShowingPetName()
	self.db.char.showPetName = not self.db.char.showPetName
    if not (self.db.char.showPetName or self.db.char.showPetLevel) then self:ShowIcon() end
	self:UpdateText()
	return self.db.char.showPetName
end

function PetInfoFu:IsShowingPetLevel()
	return self.db.char.showPetLevel
end

function PetInfoFu:ToggleIsShowingPetLevel()
	self.db.char.showPetLevel = not self.db.char.showPetLevel
    if not (self.db.char.showPetName or self.db.char.showPetLevel) then self:ShowIcon() end
	self:UpdateText()
	return self.db.char.showPetLevel
end

function PetInfoFu:IsShowingAttributes()
	return self.db.char.showAttributes
end

function PetInfoFu:ToggleIsShowingAttributes()
	self.db.char.showAttributes = not self.db.char.showAttributes
	self:UpdateTooltip()
	return self.db.char.showAttributes
end

function PetInfoFu:IsShowingCombatStats()
	return self.db.char.showCombatStats
end

function PetInfoFu:ToggleIsShowingCombatStats()
	self.db.char.showCombatStats = not self.db.char.showCombatStats
	self:UpdateTooltip()
	return self.db.char.showCombatStats
end

function PetInfoFu:IsShowingSkills()
	return self.db.char.showSkills
end

function PetInfoFu:ToggleIsShowingSkills()
	self.db.char.showSkills = not self.db.char.showSkills
	self:UpdateTooltip()
	return self.db.char.showSkills
end

function PetInfoFu:IsShowingStableMapNotes()
	return self.db.char.showStableMapNotes
end

function PetInfoFu:ToggleIsShowingStableMapNotes()
	self.db.char.showStableMapNotes = not self.db.char.showStableMapNotes
	self:UpdateTooltip()
	return self.db.char.showStableMapNotes
end

function PetInfoFu:IsShowingStabledPets()
	return self.db.char.showStabledPets
end

function PetInfoFu:ToggleIsShowingStabledPets()
	self.db.char.showStabledPets = not self.db.char.showStabledPets
	self:UpdateTooltip()
	return self.db.char.showStabledPets
end

function PetInfoFu:IsTrackingPetHistory()
	return self.db.char.trackPetHistory
end

function PetInfoFu:ToggleIsTrackingPetHistory()
	self.db.char.trackPetHistory = not self.db.char.trackPetHistory
	self:UpdateText()
	return self.db.char.trackPetHistory
end
                                          
-- init/exit functions

function PetInfoFu:OnInitialize()
  --nothing here
end

function PetInfoFu:OnEnable()
    -- update
    self:RegisterEvent("PLAYER_ENTERING_WORLD","Update")
	self:RegisterEvent("PLAYER_LEVEL_UP","Update")
    -- seems not to catch a change in demon
    self:RegisterEvent("PLAYER_PET_CHANGED","Update")
   	self:RegisterEvent("UNIT_PET","Update")
	self:RegisterEvent("UNIT_PET_EXPERIENCE","Update")
	self:RegisterEvent("UNIT_PET_TRAINING_POINTS","Update")
    
    -- local event handlers
    if self:IsHunter() then
        self:RegisterEvent("CRAFT_SHOW")
        self:RegisterEvent("PET_STABLE_CLOSED")
    	self:RegisterEvent("UNIT_HAPPINESS")
    end
    if self:IsWarlock() then
        self:GetDemonicKnowledgeRank()
        self:RegisterEvent("CHARACTER_POINTS_CHANGED")
    end
    
    self:SetIcon(BS:GetSpellIcon("Train Beast"))
    
    -- mapnotes
 	if Cartographer_Notes then
    	Cartographer_Notes:RegisterIcon("Stable", {text = L["Stable"], 
                                                   path = "Interface\\AddOns\\FuBar_PetInfoFu\\icons\\Stable"
                                                }
                                        )
        Cartographer_Notes:RegisterNotesDatabase('PetInfoFu', self.db.account.stables, self)
    end
end

function PetInfoFu:OnDisable()
    if Cartographer_Notes then
		Cartographer_Notes:UnregisterIcon("Stable")  
        Cartographer_Notes:UnregisterNotesDatabase('PetInfoFu')
    end
    self:UnregisterAllEvents()
end

-- events

function PetInfoFu:CHARACTER_POINTS_CHANGED()
    self:GetDemonicKnowledgeRank()
end

function PetInfoFu:CRAFT_SHOW()
    for i = 1, GetNumCrafts() do
        local craftName, craftSubSpellName, _, _, _, _, requiredLevel = GetCraftInfo(i)
        if craftName ~= nil then
            if self.db.char.beastTraining[craftName] == nil then
                self.db.char.beastTraining[craftName] = {}                        
            end
            if self.db.char.beastTraining[craftName][craftSubSpellName] == nil then 
                self.db.char.beastTraining[craftName][craftSubSpellName] = {
                    RequiredLevel = requiredLevel,
                }
                DEFAULT_CHAT_FRAME:AddMessage(craftName.." "..craftSubSpellName.." ".." "..requiredLevel);
            end
        end
    end
end    

function PetInfoFu:PET_STABLE_CLOSED()
    self:GetStabledPets()
    if self.db.char.trackPetHistory then
        self:TrackPetHistory()
    end
    if Cartographer_Notes and 
       self.db.char.showStableMapNotes then
        self:AddStableMapNote()
    end
end

--    seems not to catch a change of demon
--    function PetInfoFu:PLAYER_PET_CHANGED()
--        DEFAULT_CHAT_FRAME:AddMessage('PLAYER_PET_CHANGED')        
--        if self:IsWarlock() then
--            local i = 1
--            local petName = UnitName("pet")
--            local petType = UnitCreatureFamily("pet")       
--            while true do
--                local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET)
--                if not spellName then
--                    do break end
--                end
--                self:UpdateNetheredDemon(petName,spellName,spellRank)
--                i = i + 1
--            end
--        end
--        self:Update()
--    end

function PetInfoFu:UNIT_HAPPINESS()
    local happiness = GetPetHappiness()
    if happiness ~= prevHappiness then
        prevHappiness = happiness
        self:Update()
    end
end

-- functions

function PetInfoFu:GetStabledPets()
    self.db.char.stabledPets = nil
    self.db.char.stabledPets = {}
    
    for i = 1, GetNumStableSlots() do  -- i = 0 for current pet
        local _, petName, petLevel, petType, petLoyalty = GetStablePetInfo(i);
        if petName ~= nil then
            self.db.char.stabledPets[i] = {
                PetName = petName,
                PetType = petType,
                PetLoyalty = petLoyalty,
                PetLevel = petLevel
            }
        end
    end
    self:Update()
end

function PetInfoFu:TrackPetHistory()
    -- store each pet only once, with type and name as key
    -- this will get messed up if you don't name your pets 
    -- and stable a new pet of the same type or rename 
    -- a pet with the same name as an existing pet of that type
    for i = 1, GetNumStableSlots() do  -- i = 0 for current pet
        local _, petName, petLevel, petType, petLoyalty = GetStablePetInfo(i);
        if petName ~= nil then
            if self.db.char.petHistory[petType.."|"..petName] == nil then
                self.db.char.petHistory[petType.."|"..petName] = {
                    PetName = petName,
                    PetType = petType,
                    PetLevel = petLevel,
                    HunterLevel = UnitLevel("player")                
                }
            end
        end
    end
    self:Update()
end

function PetInfoFu:GetDemonicKnowledgeRank()   
     if (demonicKnowledgeTalentTab == 0 or demonicKnowledgeTalentIndex == 0) then
		for i = 1, GetNumTalentTabs() do
			local tabName = GetTalentTabInfo(i)
			if (tabName == BST["Demonology"]) then
				demonicKnowledgeTalentTab = i
				for j = 1, GetNumTalents(i) do
					local talentName, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, j)
					if talentName == BS["Demonic Knowledge"] then
						demonicKnowledgeTalentIndex = j
						--set currentRank
                        demonicKnowledgeRank = currentRank
                        
					end
				end
			end
		end
	else
		local _, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, demonicKnowledgeTalentIndex)
		-- set currentRank
        demonicKnowledgeRank = currentRank
	end
    self:Update()
end

function PetInfoFu:UpdateNetheredDemon(petName,petType,spellName,spellRank)
    if petName ~= nil then
        if self.db.char.netheredDemons[petName] == nil then
            self.db.char.netheredDemons[petName] = {}
        end
        self.db.char.netheredDemons[petName].PetName = petName
        self.db.char.netheredDemons[petName].PetType = petType
        if self.db.char.netheredDemons[petName][spellName] == nil then
            self.db.char.netheredDemons[petName][spellName] = {}
        end
        self.db.char.netheredDemons[petName][spellName] = {
            SpellName = spellName,
            SpellRank = spellRank
        }      
    end
end

function PetInfoFu:ToggleCategory(id, button)
    -- hidden is character specific data
    self.db.char.hidden[id] = not self.db.char.hidden[id]
	-- refresh in place
	self:UpdateTooltip()
end    
    
function PetInfoFu:PetInfo_GetPetExperience()
	local currentXP, totalXP = GetPetExperience()
	local toLevelXP = totalXP - currentXP
	return currentXP, totalXP, toLevelXP
end
	
function PetInfoFu:IsHunter()
	local _, englishClass = UnitClass("player")
	return englishClass == "HUNTER"
end
	
function PetInfoFu:IsWarlock()
	local _, englishClass = UnitClass("player")
	return englishClass == "WARLOCK" 
end
	
-- map notes

function PetInfoFu:AddStableMapNote()
	--get coordinates and name of stable owner
	local x, y = GetPlayerMapPosition("player")
    local stableName, _ = UnitName("target")

    -- check if we have an existing note, if not add one
	if not Cartographer_Notes:GetNearbyNote(GetRealZoneText(), x, y, 0.05, 'PetInfoFu') then
        Cartographer_Notes:SetNote(GetRealZoneText(), x, y, "Stable", "PetInfoFu",
                                   'title',  stableName
                                  )
    end
end    
    
-- menus
	
function PetInfoFu:OnMenuRequest(level, value)
	if level == 1 then
		if self:IsHunter() then
            dewdrop:AddLine(
                'text', BS["Beast Training"],
                'func', function() CastSpellByName(BS["Beast Training"])
                        end,
                'closeWhenClicked', true
            )
            dewdrop:AddLine()
		end
        -- label options
        dewdrop:AddLine(
            'text', L["Show pet's name"],
    		'func', 'ToggleIsShowingPetName',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingPetName()
        )
        dewdrop:AddLine(
            'text', L["Show pet's level"],
    		'func', 'ToggleIsShowingPetLevel',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingPetLevel()
        )
        -- tooltip options
        dewdrop:AddLine(
            'text', L["Show pet's attributes"],
    		'func', 'ToggleIsShowingAttributes',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingAttributes()
        )
        dewdrop:AddLine(
            'text', L["Show pet's combat statistics"],
    		'func', 'ToggleIsShowingCombatStats',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingCombatStats()
        )
        dewdrop:AddLine(
            'text', L["Show pet's skills"],
    		'func', 'ToggleIsShowingSkills',
 			'arg1', self,
            'closeWhenClicked', true,                
			'checked', self:IsShowingSkills()
        )
       if self:IsHunter() then
            dewdrop:AddLine(
                'text', L["Show stabled pets"],
                'func', 'ToggleIsShowingStabledPets',
                'arg1', self,
                'closeWhenClicked', true,         
                'checked', self:IsShowingStabledPets()
            )
            dewdrop:AddLine(
                'text', L["Track pet history"],
                'func', 'ToggleIsTrackingPetHistory',
                'arg1', self,
                'closeWhenClicked', true,                
                'checked', self:IsTrackingPetHistory()
            )
            dewdrop:AddLine(
                'text', L["MapNotes"],        -- stub for popup
                'arg1', self,
                'hasArrow', true,
                'value', "MAPNOTES"
            )        
        end
        if self:IsWarlock() then
            dewdrop:AddLine(
                'text', L["Show nethered demons"],
                'func', 'ToggleIsShowingStabledPets',
                'arg1', self,
                'closeWhenClicked', true,         
                'checked', self:IsShowingStabledPets()
            )        
        end        
        dewdrop:AddLine()
	end
    if level == 2 then
        if value == "MAPNOTES" then
            if self:IsHunter() then
                dewdrop:AddLine(
                    'text', L["Enable Stable mapnotes"],
                    'func', 'ToggleIsShowingStableMapNotes',
                    'arg1', self,
                    'closeWhenClicked', true,                
                    -- 'disabled', function() return not Cartographer_Notes end,
                    'checked', self:IsShowingStableMapNotes()
                )
            end    
        end
    end
end
	
-- updates    

function PetInfoFu:AddStabledPets()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_STABLED"], 'columns', 2,
                                           'text', L["CATEGORY_PET_STABLED"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_STABLED"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_STABLED"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_STABLED"]] then
        local petsFound            
        for k, v in pairs(self.db.char.stabledPets) do  
            tooltipLine:AddLine(
                'text', v.PetName..' ('..v.PetType..')',
                'text2', v.PetLevel,
                'text2R', 1, 'text2G', 1, 'text2B', 0
            )              
            petsFound = true
        end
        if not petsFound then
                tooltipLine:AddLine(
                    'text', L["Please visit a stable to initialise."]
                )
        end
    end
end
	
function PetInfoFu:AddPetHistory()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_HISTORY"], 'columns', 2,
                                           'text', L["CATEGORY_PET_HISTORY"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_HISTORY"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_HISTORY"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_HISTORY"]] then
        -- create a sorted table of pets
        sortedPets = {}
        for k,v in pairs(self.db.char.petHistory) do
            if v.HunterLevel == nil then
                v.HunterLevel = UnitLevel("player")
                self.db.char.petHistory[k].HunterLevel = v.HunterLevel
            end
            table.insert(sortedPets, v)
        end
        table.sort(sortedPets, function (a, b) return a.HunterLevel < b.HunterLevel end)
        -- list sorted pets
        for i = 1, #sortedPets do  
            tooltipLine:AddLine(
                'text', sortedPets[i].PetName..' ('..sortedPets[i].PetType..')',
                'text2', sortedPets[i].PetLevel..' ('..sortedPets[i].HunterLevel..')',
                'text2R', 1, 'text2G', 1, 'text2B', 0
            )              
        end
    end
end
 
function PetInfoFu:ShowSkills()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_SKILLS"], 'columns', 2,
                                           'text', L["CATEGORY_PET_SKILLS"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_SKILLS"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] then
        local petName = UnitName("pet")    
        local petType = UnitCreatureFamily("pet")
        local i = 1
        while true do
            local spellName, spellRank = GetSpellName(i, BOOKTYPE_PET)
            if not spellName then
                do break end
            end
            -- check if higher skill is available
            if self:IsHunter() then
                if self.db.char.beastTraining[spellName] ~= nil then
                    if self.db.char.beastTraining[spellName][spellRank] ~= nil then
                        local rankNo = string.sub(spellRank,string.find(spellRank,' ')+1)
                        local nextRank = string.sub(spellRank,1,string.find(spellRank,' '))..rankNo + 1
                        if self.db.char.beastTraining[spellName][nextRank] ~= nil then
                            local requiredLevel = self.db.char.beastTraining[spellName][nextRank].RequiredLevel
                            if requiredLevel > UnitLevel("pet") then 
                                spellRank = spellRank..' ('..crayon:Red(format(requiredLevel))..')'
                            else
                                spellRank = spellRank..' ('..format(requiredLevel)..')'
                            end
                        end
                    end
                end
            end
            if self:IsWarlock() then
                self:UpdateNetheredDemon(petName,petType,spellName,spellRank)
            end
            tooltipLine:AddLine(
                'text', spellName,
                'text2', spellRank,
                'text2R', 1, 'text2G', 1, 'text2B', 0
            )
            i = i + 1
        end
     end
end
 
function PetInfoFu:AddNetheredDemons()
    local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_NETHER"], 'columns', 2,
                                           'text', L["CATEGORY_PET_NETHER"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_NETHER"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                              			   'child_indentation', 20,                                           
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_NETHER"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
    )        
    if not self.db.char.hidden[L["CATEGORY_PET_NETHER"]] then
        -- create a sorted table of pets
        local sortedPets = {}
        for k,v in pairs(self.db.char.netheredDemons) do
             table.insert(sortedPets, v)
        end
        table.sort(sortedPets, function (a, b) return a.PetType < b.PetType end)
        -- list sorted pets
        for i = 1, #sortedPets do
            if sortedPets[i].PetName ~= UnitName('pet') then
                local tooltipLine = tablet:AddCategory('id', sortedPets[i].PetName, 'columns', 2,
                                                       'text', sortedPets[i].PetName..' ('..sortedPets[i].PetType..')',
                                                       'func', 'ToggleCategory', 'arg1', self, 'arg2', sortedPets[i].PetName,
                                                       'indentation', 10,
                                                       'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                       'child_indentation', 25,
                                                       'showWithoutChildren', true,
                                                       'hideBlankLine', true,                                           
                                                       'checked', true, 'hasCheck', true, 'checkIcon',
                                                       self.db.char.hidden[sortedPets[i].PetName] and 
                                                       'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
                )        
                if not self.db.char.hidden[sortedPets[i].PetName] then
                    local sortedSkills = {}
                    for k,v in pairs(self.db.char.netheredDemons[sortedPets[i].PetName]) do
                        if v.SpellName ~= nil then 
                            table.insert(sortedSkills, v)
                        end
                    end
                    table.sort(sortedSkills, function (a, b) return a.SpellName < b.SpellName end)
                
                    for j = 1, #sortedSkills do
                        tooltipLine:AddLine(
                            'text', sortedSkills[j].SpellName,
                            'text2', sortedSkills[j].SpellRank,
                            'text2R', 1, 'text2G', 1, 'text2B', 0
                        ) 
                    end                    
                end
            end
        end
    end
end
 
function PetInfoFu:OnTooltipUpdate()

    if HasPetUI() then

        if self:IsHunter() then
	        tablet:SetHint(L["PI_TOOLTIP_HINT_HUNTER"])
        else
            tablet:SetHint(L["PI_TOOLTIP_HINT"])       
        end

        
        local toolTip = tablet:AddCategory(
            'columns', 2,
            'child_textR', 0,
            'child_textG', 1,
            'child_textB', 0,
            'child_text2R', 1,
            'child_text2G', 1,
            'child_text2B', 0
        )
        
        local petName = UnitName("pet")
        toolTip:AddLine(
            'text', L["PI_PETNAME"],
            'text2', petName
        )        
        local petLevel = UnitLevel("pet")
        toolTip:AddLine(
            'text', L["PI_PETLEVEL"],
            'text2', petLevel
        )
        if self:IsHunter() then
            -- general pet info            
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_INFO"], 'columns', 2,
                                                   'text', L["CATEGORY_PET_INFO"],
                                                   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_INFO"],
                                                   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                   'showWithoutChildren', true,
                                                   'checked', true, 'hasCheck', true, 'checkIcon',
                                                   self.db.char.hidden[L["CATEGORY_PET_INFO"]] and 
                                                   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )
            if not self.db.char.hidden[L["CATEGORY_PET_INFO"]] then
            
                local petType = UnitCreatureFamily("pet")
                
                tooltipLine:AddLine(
                    'text', L["PI_TYPE"],
                    'text2', petType,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine(
                    'text', L["PI_FOOD"],
                    'text2', BuildListString(GetPetFoodTypes()),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine(
                    'text', L["PI_LOYALTY"],
                    'text2', GetPetLoyalty(),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- happiness info
                local happiness, damagePercentage, loyaltyRate = GetPetHappiness()
                local happyText = ''
                local loyaltyText = ''
                local rHappy = 1 
                local gHappy = 1
                local bHappy = 0

                if happiness == 1 then
                    happyText = L["Unhappy"]
                    rHappy = 1
                    gHappy = 0
                elseif happiness == 2 then
                    happyText = L["Content"]              
                elseif happiness == 3 then
                    happyText = L["Happy"]
                    rHappy = 0
                    gHappy = 1
                end
                
                if loyaltyRate > 0 then
                    loyaltyText = L["Gaining Loyalty"]
                else
                    loyaltyText = L["Losing Loyalty"]
                end
                tooltipLine:AddLine(
                    'text', L["PI_HAPPINESS"],
                    'text2', happyText,
                    'text2R', rHappy, 'text2G', gHappy, 'text2B', bHappy
                )            
                tooltipLine:AddLine(
                    'text', L["PI_DAMAGEPERCT"],
                    'text2', damagePercentage..'%',
                    'text2R', rHappy, 'text2G', gHappy, 'text2B', bHappy
                )
                tooltipLine:AddLine(
                    'text', L["PI_LOYALTYRATE"],
                    'text2', loyaltyText,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- experience info          
                local currentXP, totalXP, toLevelXP = self:PetInfo_GetPetExperience()
                local currentXPPercent = currentXP / totalXP * 100
                local toLevelXPPercent = toLevelXP / totalXP * 100            
                
                if petLevel < UnitLevel("player") then
                    tooltipLine:AddLine(
                        'text', L["PI_CURRENT_XP"], 
                        'text2', format(L["PI_PERCENT_FORMAT"], currentXP, currentXPPercent),
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                    tooltipLine:AddLine(
                        'text', L["PI_NEEDED_XP"], 
                        'text2', format(L["PI_PERCENT_FORMAT"], toLevelXP, toLevelXPPercent),
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                    tooltipLine:AddLine(
                        'text', L["PI_TOTAL_XP"], 
                        'text2', totalXP,
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                else
                    tooltipLine:AddLine(
                        'text', L["PI_NOLEVEL"]
                    )
                end
                tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()
                
                -- training info
                local totalTP, usedTP = GetPetTrainingPoints()
                local freeTP = totalTP - usedTP            
                tooltipLine:AddLine(
                    'text', L["PI_TRAINING_POINTS"], 
                    'text2', freeTP,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
            end
        end
        
        if self:IsWarlock() then
            -- demonic knowledge
            -- skip section if not talented in demonic knowledge
            if demonicKnowledgeRank > 0 then
                local tooltipLine = tablet:AddCategory('id', L["CATEGORY_DEMONIC_KNOWLEDGE"], 'columns', 2,
                                                       'text', L["CATEGORY_DEMONIC_KNOWLEDGE"],
                                                       'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_DEMONIC_KNOWLEDGE"],                                                   
                                                       'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                       'showWithoutChildren', true,
                                                       'checked', true, 'hasCheck', true, 'checkIcon',
                                                       self.db.char.hidden[L["CATEGORY_DEMONIC_KNOWLEDGE"]] and 
                                                       'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'                                                  
                )
            
                if UnitExists("pet") then
                    local demonSta, demonInt
                    _, demonSta = UnitStat("pet", 3) -- stamina
                    _, demonInt = UnitStat("pet", 4) -- intelligence
                    tooltipLine:AddLine(
                        'text', L["PI_BONUSDAMAGE"]..' ('..demonicKnowledgeRank..')',
                        'text2', floor((demonSta + demonInt) * 0.05 * demonicKnowledgeRank),
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                else
                    tooltipLine:AddLine(
                        'text', L["PI_NODEMON"]
                    )            
                end
            end
        end
        
        -- combat info
        if self.db.char.showCombatStats then        
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_COMBAT_INFO"], 'columns', 2,
                                           'text', L["CATEGORY_PET_COMBAT_INFO"],
                                           'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_COMBAT_INFO"],
                                           'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                           'showWithoutChildren', true,
                                           'checked', true, 'hasCheck', true, 'checkIcon',
                                           self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] and 
                                           'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )
            if not self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] then

                local baseArmor, effectiveArmor, armor, posArmorBuff, negArmorBuff = UnitArmor("pet");
                if posArmorBuff - negArmorBuff < 0 then
                    effectiveArmor = crayon:Red(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Red("-"..format(baseArmor - effectiveArmor))..')'
                elseif baseArmor < effectiveArmor then 
                    effectiveArmor = crayon:Green(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Green("+"..format(effectiveArmor - baseArmor))..')'
                end
-- DEFAULT_CHAT_FRAME:AddMessage(baseArmor.." "..effectiveArmor.." "..armor.." "..posArmorBuff.." ".." "..negArmorBuff)                

                tooltipLine:AddLine(
                    'text', L["PI_ARMOR_CLASS"], 
                    'text2', effectiveArmor,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
              
                local atkBase, atkMod = UnitAttackBothHands("pet");        
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_SKILL"],
                    'text2', atkBase,
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )

                local atkPowerbase, atkPowerPosBuff, atkPowerNegBuff = UnitAttackPower("pet");
                if atkPowerPosBuff - atkPowerNegBuff < 0 then
                    atkPowerbase = crayon:Red(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Red("-"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
                elseif atkPowerPosBuff - atkPowerNegBuff > 0 then
                    atkPowerbase = crayon:Green(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Green("+"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
                end
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_POWER"],
                    'text2', atkPowerbase,
                    'text2R', 1, 'text2G', 1, 'text2B', 0                
                )
                
                local atkSpeed = UnitAttackSpeed("pet")       
                tooltipLine:AddLine(
                    'text', L["PI_ATTACK_SPEED"], 
                    'text2', format(L["PI_AS_FORMAT"], atkSpeed),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )

                local lowDmg, hiDmg, offlowDmg, offhiDmg, posBuff, negBuff, percentMod = UnitDamage("pet");
                local listedDPS = (((lowDmg + hiDmg) * .5 + posBuff + negBuff) * percentMod) / atkSpeed;
                tooltipLine:AddLine(
                    'text', L["PI_DPS_LISTED"], 
                    'text2', format(L["PI_DPS_FORMAT"], listedDPS),
                    'text2R', 1, 'text2G', 1, 'text2B', 0
                )
            end
        end
        
        -- show stats info if showing
        if self.db.char.showAttributes then        
            local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_ATTRIBUTES"], 'columns', 2,
                                                   'text', L["CATEGORY_PET_ATTRIBUTES"],
                                                   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_ATTRIBUTES"],
                                                   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
                                                   'showWithoutChildren', true,
                                                   'checked', true, 'hasCheck', true, 'checkIcon',
                                                   self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] and 
                                                   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
            )        
            if not self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] then         
                for i = 1, NUM_PET_STATS, 1 do                  
                    local attrbStr = L["SPELL_STAT"..format(i-1).."_NAME"]                   
                    local baseStat, effectiveStat, posBuff, negBuff = UnitStat("pet", i)
            
                    if posBuff == 0 and negBuff == 0 then
                        baseStat = effectiveStat
                    else 
                        -- if there are any negative buffs then show the main number in red even if there are
                        -- positive buffs. otherwise show in green.
                        if negBuff < 0 then
                            baseStat = crayon:Red(effectivestat)..' ('..baseStat.."-"..crayon:Red(negBuff+posBuff)..')'
                        else
                            baseStat = crayon:Green(effectiveStat)..' ('..baseStat-posBuff.."+"..crayon:Green(posBuff)..')'
                        end
                    end
                    tooltipLine:AddLine(
                        'text', attrbStr,
                        'text2', baseStat,
                        'text2R', 1, 'text2G', 1, 'text2B', 0
                    )
                end
            end
        end
        
        -- show skills info if showing
        if self.db.char.showSkills then        
            self:ShowSkills()
        end        

        -- add other demons if showing
        if self:IsWarlock() and self.db.char.showStabledPets then          
            self:AddNetheredDemons()
        end   
        
        -- add other pets if showing
        if self:IsHunter() and self.db.char.showStabledPets then          
            self:AddStabledPets()
        end   
        -- add pet history
        if self:IsHunter() and self.db.char.trackPetHistory then
            self:AddPetHistory()
        end           
    else
        -- no pet found
        local toolTip = tablet:AddCategory(
            'columns', 1,
            'child_textR', 0,
            'child_textG', 1,
            'child_textB', 0
        )      
        toolTip:AddLine(
            'text', L["PI_NOPET"]
        )
        -- add stabled pets
        if self:IsHunter() and self.db.char.showStabledPets then
            self:AddStabledPets()
        end   
        -- add pet history
        if self:IsHunter() and self.db.char.trackPetHistory then
            self:AddPetHistory()
        end           
    end
end

function PetInfoFu:UpdateText()
    if  self:IsTextShown() then
        if self.db.char.showPetName or self.db.char.showPetLevel then
            if UnitName("pet") ~= nil then
                local petName = UnitName("pet")
                local petLevel = UnitLevel("pet")
                local petIcon = GetPetIcon()
                local happiness = GetPetHappiness()
                local label = ''
                
                if petName ~= nil and petLevel ~= nil then  
                
                    if petName == UNKNOWNOBJECT then
                        petName = L["UO"]
                    end
                
                    if petIcon ~= nil then
                        self:SetIcon(petIcon)        
                    end
                    
                    if self.db.char.showPetName then                   
                        if self.db.char.showPetLevel then                        
                            label = petName..' ('..petLevel..')'
                        else
                            label = petName
                        end
                    else
                        if self.db.char.showPetLevel then    
                            label = petLevel
                        else
                                                
                        end 
                    end
                    
                    if happiness == nil then             -- non hunter pet
                        self:SetText(label)
                    elseif happiness == 1 then
                        self:SetText(crayon:Red(label))    -- unhappy
                    elseif happiness == 2 then
                        self:SetText(crayon:Yellow(label)) -- content              
                    elseif happiness == 3 then
                        self:SetText(crayon:Green(label))  -- happy
                    end
                else
                    -- pet name and level are nil which means there shouldn't have been a pet ui in the first place
                    self:SetText(L["PI_NOPET"])
                    self:SetIcon(BS:GetSpellIcon("Train Beast"))
                end
            else 
                -- no pet ui found
                self:SetText(L["PI_NOPET"])
                self:SetIcon(BS:GetSpellIcon("Train Beast"))
            end
        else
            if UnitName("pet") ~= nil then
                local petIcon = GetPetIcon()
                if petIcon ~= nil then
                    self:SetIcon(petIcon)        
                end
            end
            self:SetText("")                    
        end
    end
end	

-- event handlers

function PetInfoFu:OnClick()
    if self:IsHunter() then
        CastSpellByName(BS["Beast Training"])
    end
end
